package dev.flyfish.framework.context;

import org.springframework.stereotype.Component;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

/**
 * 基于ThreadLocal的时间仓库，提高线程性能
 *
 * @author wybab
 */
@Component
public class DateContext {

    private static DateContext instance;
    private ThreadLocal<DateFormat> formatThreadLocal = new ThreadLocal<>();
    private ThreadLocal<Map<String, DateFormat>> mapThreadLocal = new ThreadLocal<>();
    private static final ZoneId zoneId = ZoneId.of("Asia/Shanghai");

    public DateContext() {
        instance = this;
    }

    public static DateContext getInstance() {
        return instance;
    }

    public static LocalDateTime ofMillis(long millis) {
        Instant instant = Instant.ofEpochMilli(millis);
        return LocalDateTime.ofInstant(instant, zoneId);
    }

    public static long toMillis(LocalDateTime localDateTime) {
        return localDateTime.atZone(zoneId).toInstant().toEpochMilli();
    }

    public static long distance(LocalDateTime start, LocalDateTime end) {
        return toMillis(end) - toMillis(start);
    }

    public static DateFormat simpleFormat() {
        DateFormat format = instance.formatThreadLocal.get();
        if (format == null) {
            format = new SimpleDateFormat("yyyy-MM-dd");
            instance.formatThreadLocal.set(format);
        }
        return format;
    }

    /**
     * 内置的解析，避免使用
     *
     * @param date 日期
     * @return 结果
     */
    public static Date parse(Object date) {
        if (date instanceof String) {
            try {
                return simpleFormat().parse((String) date);
            } catch (ParseException e) {
                e.printStackTrace();
            }
        }
        return null;
    }

    /**
     * 内置解析
     *
     * @param date    日期
     * @param pattern 格式
     * @return 结果
     */
    public static Date parse(Object date, String pattern) {
        if (date instanceof String) {
            try {
                return dateFormat(pattern).parse((String) date);
            } catch (ParseException e) {
                e.printStackTrace();
            }
        }
        return null;
    }

    public static DateFormat dateFormat(String pattern) {
        return instance.getDateFormat(pattern);
    }

    public static String formatDate(Date date, String pattern) {
        return dateFormat(pattern).format(date);
    }

    public DateFormat getDateFormat(String pattern) {
        Map<String, DateFormat> formatMap = mapThreadLocal.get();
        if (null == formatMap) {
            formatMap = new HashMap<>();
            mapThreadLocal.set(formatMap);
        }
        DateFormat format = formatMap.compute(pattern, (key, value) ->
                value == null ? new SimpleDateFormat(key) : value);
        return format;
    }
}
