package dev.flyfish.framework.mongodb.builder;

import dev.flyfish.framework.context.DateContext;
import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.domain.base.Qo;
import com.mongodb.BasicDBList;
import lombok.Getter;
import lombok.Setter;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.apache.commons.lang3.ClassUtils;
import org.bson.types.ObjectId;
import org.springframework.beans.BeanUtils;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.util.ObjectUtils;

import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.function.BiFunction;
import java.util.function.Supplier;
import java.util.stream.Collectors;

/**
 * 查询构建器
 *
 * @author wybab
 */
public final class CriteriaBuilder<T extends Domain> {

    private final Map<String, BiFunction<Criteria, Object, Criteria>> functionMap = new HashMap<>();
    private final Map<String, String> keyMapper = new HashMap<>();
    private final List<Supplier<Criteria>> criterias = new ArrayList<>();
    private Qo<T> qo;

    /**
     * 构造器，接受一个qo
     *
     * @param qo  查询实体
     * @param <T> 泛型，实体
     * @return 结果
     */
    public static <T extends Domain> CriteriaBuilder<T> accept(Qo<T> qo) {
        CriteriaBuilder<T> builder = new CriteriaBuilder<>();
        builder.qo = qo;
        return builder;
    }

    /**
     * 创建criteria列表
     *
     * @param criteria 多个查询
     * @return 结果
     */
    public static BasicDBList createCriteriaList(Criteria... criteria) {
        BasicDBList bsonList = new BasicDBList();
        for (Criteria c : criteria) {
            bsonList.add(c.getCriteriaObject());
        }
        return bsonList;
    }

    /**
     * 添加策略
     *
     * @param key      键
     * @param function 方法
     * @param field    域的键值
     * @return 结果
     */
    public CriteriaBuilder<T> with(String key, String field, BiFunction<Criteria, Object, Criteria> function) {
        this.functionMap.putIfAbsent(key, function);
        keyMapper.putIfAbsent(key, field);
        return this;
    }

    /**
     * 添加策略
     *
     * @param key      键
     * @param function 方法
     * @return 结果
     */
    public CriteriaBuilder<T> with(String key, BiFunction<Criteria, Object, Criteria> function) {
        this.functionMap.putIfAbsent(key, function);
        return this;
    }

    /**
     * 添加策略
     *
     * @param key 键
     * @return 结果
     */
    public CriteriaBuilder<T> with(String... key) {
        Arrays.stream(key).forEach(item -> this.functionMap.putIfAbsent(item, Criteria::is));
        return this;
    }

    /**
     * 添加自定义条件
     *
     * @param function 函数
     * @return 结果
     */
    public CriteriaBuilder<T> with(Supplier<Criteria> function) {
        this.criterias.add(function);
        return this;
    }

    /**
     * 批量导入规则
     *
     * @param criteriaDescriptors 规则集合
     * @return 结果
     */
    public CriteriaBuilder<T> with(List<CriteriaDescriptor> criteriaDescriptors) {
        criteriaDescriptors.forEach(criteriaDescriptor ->
                this.with(criteriaDescriptor.getKey(), criteriaDescriptor.getMapper()));
        return this;
    }

    public Criteria build() {
        if (!MapUtils.isEmpty(functionMap)) {
            // 键集合
            Set<String> keySet = functionMap.keySet();
            // 建立查询器
            List<Criteria> criteria = Arrays.stream(BeanUtils.getPropertyDescriptors(qo.getClass()))
                    .filter(propertyDescriptor -> keySet.contains(propertyDescriptor.getName()))
                    .map(propertyDescriptor -> {
                        try {
                            String key = propertyDescriptor.getName();
                            // 值
                            Object value = propertyDescriptor.getReadMethod().invoke(qo);
                            // 值非空予以过滤
                            if (!ObjectUtils.isEmpty(value)) {
                                String field = determineField(key, propertyDescriptor);
                                return functionMap.getOrDefault(propertyDescriptor.getName(), Criteria::is).apply(
                                        Criteria.where(field), fixValue(field, value));
                            }
                        } catch (IllegalAccessException | InvocationTargetException e) {
                            e.printStackTrace();
                        }
                        return null;
                    })
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
            // 添加自定义criteria
            if (CollectionUtils.isNotEmpty(criterias)) {
                for (Supplier<Criteria> builder : criterias) {
                    Criteria built = builder.get();
                    if (null != built) {
                        criteria.add(built);
                    }
                }
            }
            return combine(criteria);
        }
        return null;
    }

    /**
     * 修正查询最终值，主要是ObjectId
     *
     * @return 结果
     */
    @SuppressWarnings("unchecked")
    private Object fixValue(String field, Object value) {
        if (field.contains(".$id")) {
            if (value instanceof Collection) {
                Collection<Object> collection = (Collection<Object>) value;
                return collection.stream()
                        .map(item -> (item instanceof ObjectId ? item : createObjectId((String) item)))
                        .collect(Collectors.toList());
            } else {
                return createObjectId((String) value);
            }
        }
        return value;
    }

    private Object createObjectId(String value) {
        try {
            return new ObjectId(value);
        } catch (Exception e) {
            return value;
        }
    }

    private String determineField(String key, PropertyDescriptor propertyDescriptor) {
        String field = keyMapper.getOrDefault(key, key);
        if (ClassUtils.isAssignable(propertyDescriptor.getPropertyType(), Collection.class)) {
            field = field.replace(".$id", "");
        }
        return field;
    }

    /**
     * 整合最后的结果
     *
     * @param criteria 结果
     * @return 返回
     */
    private Criteria combine(List<Criteria> criteria) {
        if (criteria.size() == 0) {
            return new Criteria();
        }
        if (criteria.size() == 1) {
            return criteria.get(0);
        }
        return new Criteria().andOperator(criteria);
    }

    /**
     * 匹配策略
     */
    public interface Builders {

        // 模糊匹配
        BiFunction<Criteria, Object, Criteria> LIKE = (criteria, o) -> criteria.regex((String) o);

        // 精确匹配
        BiFunction<Criteria, Object, Criteria> IS = Criteria::is;

        // 不等
        BiFunction<Criteria, Object, Criteria> NE = Criteria::ne;

        // 包含匹配
        BiFunction<Criteria, Object, Criteria> IN = (criteria, o) -> o instanceof Collection ? criteria.in(((Collection) o)) : criteria.in(o);

        // 不包含匹配
        BiFunction<Criteria, Object, Criteria> NIN = (criteria, o) -> o instanceof Collection ? criteria.nin(((Collection) o)) : criteria.nin(o);

        // 不为空
        BiFunction<Criteria, Object, Criteria> NOT_NULL = ((criteria, o) -> criteria.ne(null));

        // 范围匹配
        BiFunction<Criteria, Object, Criteria> RANGE = (criteria, o) -> {
            if (o instanceof List) {
                List list = (List) o;
                return criteria.gte(list.get(0)).lte(list.get(1));
            }
            return criteria;
        };

        // 时间起始
        BiFunction<Criteria, Object, Criteria> DATE_GTE = (criteria, o) -> criteria.gte(DateContext.parse(o));

        // 时间中止
        BiFunction<Criteria, Object, Criteria> DATE_LTE = (criteria, o) -> criteria.lte(DateContext.parse(o));

        // 日期范围包含
        @SuppressWarnings("unchecked")
        BiFunction<Criteria, Object, Criteria> DATE_RANGE = (criteria, o) -> {
            if (o instanceof List) {
                List<String> list = (List<String>) o;
                if (list.size() == 2) {
                    criteria.gte(DateContext.parse(list.get(0)))
                            .lte(DateContext.parse(list.get(1)));
                }
            } else if (o instanceof String[]) {
                String[] array = (String[]) o;
                if (array.length == 2) {
                    criteria.gte(DateContext.parse(array[0]))
                            .lte(DateContext.parse(array[1]));
                }
            }
            return criteria;
        };
    }

    /**
     * 查询描述器
     */
    @Getter
    @Setter
    protected static class CriteriaDescriptor {

        private String key;

        private BiFunction<Criteria, Object, Criteria> mapper;
    }
}
