package dev.flyfish.framework.r2dbc.config.callback;

import dev.flyfish.framework.auditor.callback.ReactiveAuditorCallback;
import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.domain.base.Qo;
import dev.flyfish.framework.r2dbc.metadata.R2dbcMetadataManager;
import dev.flyfish.framework.r2dbc.metadata.R2dbcTableMetadata;
import dev.flyfish.framework.r2dbc.metadata.visitor.MetadataHandler;
import dev.flyfish.framework.r2dbc.metadata.visitor.R2dbcTableMetadataVisitor;
import dev.flyfish.framework.repository.DefaultReactiveRepository;
import org.reactivestreams.Publisher;
import org.springframework.beans.factory.ObjectProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mapping.PersistentEntity;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.mapping.OutboundRow;
import org.springframework.data.r2dbc.mapping.event.AfterConvertCallback;
import org.springframework.data.r2dbc.mapping.event.AfterSaveCallback;
import org.springframework.data.r2dbc.mapping.event.BeforeConvertCallback;
import org.springframework.data.relational.core.sql.SqlIdentifier;
import org.springframework.data.util.CastUtils;
import org.springframework.data.util.Lazy;
import org.springframework.lang.NonNull;
import reactor.core.publisher.Mono;

import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

public class ReferenceR2dbcCallback extends ReactiveAuditorCallback implements BeforeConvertCallback<Domain>, AfterConvertCallback<Domain>, AfterSaveCallback<Domain>, RepositoryProvider, Supplier<R2dbcEntityOperations> {

    private final R2dbcMetadataManager r2dbcMetadataManager;
    private final MetadataHandler fetchHandler;
    private final MetadataHandler writeHandler;
    private Lazy<R2dbcEntityOperations> operations;
    private Lazy<Map<Class<?>, DefaultReactiveRepository<?>>> repositories;


    public ReferenceR2dbcCallback(R2dbcMetadataManager r2dbcMetadataManager) {
        this.r2dbcMetadataManager = r2dbcMetadataManager;
        this.fetchHandler = new AssociationFetcher(this, this);
        this.writeHandler = new AssociationWriter(this, this);
    }

    /**
     * 实体回调方法，用于在域对象转换为持久化对象之前调用。该方法可以返回相同的域对象，也可以返回经过修改的域对象。
     *
     * @param entity 要保存的领域对象。
     * @param table  表名。
     * @return 要持久化的领域对象。
     */
    @Override
    @NonNull
    public Publisher<Domain> onBeforeConvert(@NonNull Domain entity, @NonNull SqlIdentifier table) {
        return audit(entity);
    }

    /**
     * 实体回调方法，用于在域对象从数据库行数据转换为对象之后调用。该方法可以返回相同的域对象，也可以返回经过修改的域对象。
     *
     * @param entity 领域对象（转换后的结果）。
     * @param table  表名。
     * @return 从数据库行读取得到的域对象（可能经过修改）。
     */
    @Override
    @NonNull
    public Publisher<Domain> onAfterConvert(@NonNull Domain entity, @NonNull SqlIdentifier table) {
        return Mono.deferContextual(ctx -> {
            Optional<Qo<? extends Domain>> query = ctx.getOrEmpty(Qo.KEY);
            query.ifPresent(qo -> entity.setCurrentUser(qo.getUser()));
            // 判断是否需要获取关联信息，如果没有qo，默认需要拉取
            boolean fetchRefs = query.map(Qo::isFetchRef).orElse(true);
            if (fetchRefs) {
                R2dbcTableMetadata metadata = r2dbcMetadataManager.getMetadata(entity.getClass());
                R2dbcTableMetadataVisitor visitor = new R2dbcTableMetadataVisitor(entity, fetchHandler);
                metadata.visit(visitor);
                return visitor.getResult();
            }
            return Mono.just(entity);
        });
    }

    /**
     * 保存完成后的回调，用于保存额外的关联对象
     *
     * @param entity      已经保存的实体，带有id信息
     * @param outboundRow {@link OutboundRow} 输出行，包含实体的完整内容 {@code entity}.
     * @param table       表名
     * @return 当前已经实例化的对象
     */
    @Override
    @NonNull
    public Publisher<Domain> onAfterSave(@NonNull Domain entity, @NonNull OutboundRow outboundRow, @NonNull SqlIdentifier table) {
        Class<? extends Domain> entityClass = entity.getClass();
        // 遍历fields，找到要注入的数据
        R2dbcTableMetadata metadata = r2dbcMetadataManager.getMetadata(entityClass);
        if (metadata.hasAssociation()) {
            R2dbcTableMetadataVisitor visitor = new R2dbcTableMetadataVisitor(entity, writeHandler);
            metadata.visit(visitor);
            return visitor.getResult();
        }
        return Mono.just(entity);
    }

    @Override
    public <T extends Domain, R extends DefaultReactiveRepository<T>> R get(PersistentEntity<?, ?> entity) {
        return CastUtils.cast(repositories.get().get(entity.getType()));
    }

    @Autowired
    public void setRepositories(ObjectProvider<DefaultReactiveRepository<?>> repositories) {
        this.repositories = Lazy.of(() -> repositories.stream()
                .collect(Collectors.toMap(repo -> repo.getEntityInformation().getJavaType(), Function.identity())));
    }

    @Autowired
    public void setR2dbcEntityOperations(ObjectProvider<R2dbcEntityOperations> operations) {
        this.operations = Lazy.of(operations::getIfAvailable);
    }


    /**
     * 获取实体操作
     */
    @Override
    public R2dbcEntityOperations get() {
        return operations.get();
    }

}
