package dev.flyfish.framework.r2dbc.metadata.impl;

import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.r2dbc.metadata.R2dbcMetadataManager;
import dev.flyfish.framework.r2dbc.metadata.R2dbcTableMetadata;
import dev.flyfish.framework.r2dbc.metadata.reference.field.FieldAccessor;
import dev.flyfish.framework.relational.mapping.Association;
import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.springframework.core.ResolvableType;
import org.springframework.core.annotation.AnnotatedElementUtils;
import org.springframework.data.annotation.Reference;
import org.springframework.data.mapping.PersistentEntity;
import org.springframework.data.mapping.PersistentProperty;
import org.springframework.data.r2dbc.mapping.R2dbcMappingContext;

import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.function.Consumer;
import java.util.stream.StreamSupport;

/**
 * 简单的元数据管理器
 *
 * @author wangyu
 * 基于线程安全的反射方法进行加载并持有
 */
@RequiredArgsConstructor
public class SimpleR2dbcMetadataManager implements R2dbcMetadataManager {

    private final R2dbcMappingContext r2dbcMappingContext;
    private final ConcurrentMap<Class<? extends Domain>, R2dbcTableMetadata> METADATA_STORE = new ConcurrentHashMap<>();

    /**
     * 获取元数据
     *
     * @param entityClass 实体类名
     * @return 结果
     */
    @Override
    public R2dbcTableMetadata getMetadata(Class<? extends Domain> entityClass) {
        return METADATA_STORE.computeIfAbsent(entityClass, this::computeMetadata);
    }

    /**
     * 计算元数据信息
     *
     * @param entityClass 实体类
     * @return 结果
     */
    private R2dbcTableMetadata computeMetadata(Class<? extends Domain> entityClass) {
        PersistentEntity<?, ?> entity = r2dbcMappingContext.getRequiredPersistentEntity(entityClass);
        MetadataCreator creator = new MetadataCreator(entity);
        FieldUtils.getFieldsListWithAnnotation(entityClass, Association.class).forEach(creator);
        return creator.metadata;
    }

    private class MetadataCreator implements Consumer<Field> {

        private final PersistentEntity<?, ?> entity;

        private final R2dbcTableMetadata metadata;

        private MetadataCreator(PersistentEntity<?, ?> entity) {
            this.entity = entity;
            this.metadata = new R2dbcTableMetadata(entity);
        }

        /**
         * 处理所有关联属性
         *
         * @param field 属性永远都不为空
         */
        @Override
        public void accept(Field field) {
            Association association = AnnotatedElementUtils.findMergedAnnotation(field, Association.class);
            if (null == association) return;
            // 获取字段访问器，只支持标准的bean
            FieldAccessor accessor = FieldAccessor.of(entity.getType(), field);
            if (!accessor.isAvailable()) return;
            // 获取具体类型
            Class<?> type = field.getType();
            if (ClassUtils.isAssignable(type, Domain.class)) {
                // 目标类型
                PersistentEntity<?, ?> targetEntity = r2dbcMappingContext.getRequiredPersistentEntity(type);
                // 多对多关联
                if (StringUtils.isNotBlank(association.relationTable())) {
                    metadata.addRelation(targetEntity, association, accessor);
                    return;
                }
                // 一对一关联
                if (StringUtils.isNotBlank(association.field())) {
                    // 获取内部属性
                    PersistentProperty<?> property = entity.getRequiredPersistentProperty(association.field());
                    // 添加内部关联
                    metadata.addAssociation(targetEntity, property, accessor, true);
                } else if (StringUtils.isNotBlank(association.foreignField())) {
                    // 获取外部属性
                    PersistentProperty<?> property = targetEntity.getRequiredPersistentProperty(association.foreignField());
                    // 添加外部关联
                    metadata.addAssociation(targetEntity, property, accessor, false);
                } else {
                    // 都为空，从当前对象查找目标类型的reference，并自动添加关联
                    Optional<PersistentProperty<?>> inner = findInnerKey(type);
                    if (inner.isPresent()) {
                        metadata.addAssociation(targetEntity, inner.get(), accessor, true);
                    } else {
                        findForeignKey(targetEntity).ifPresent(property -> metadata.addAssociation(targetEntity, property, accessor, false));
                    }
                }
            } else if (ClassUtils.isAssignable(type, Collection.class)) {
                // 获取真实的类型
                Class<?> entityType = ResolvableType.forField(field).resolveGeneric(0);
                if (null == entityType) return;
                // 目标类型
                PersistentEntity<?, ?> targetEntity = r2dbcMappingContext.getRequiredPersistentEntity(entityType);
                // 多对多关联
                if (StringUtils.isNotBlank(association.relationTable())) {
                    metadata.addRelation(targetEntity, association, accessor);
                    return;
                }
                // 一对多关联，仅支持外部字段
                if (StringUtils.isNotBlank(association.foreignField())) {
                    // 获取外部属性
                    PersistentProperty<?> property = targetEntity.getRequiredPersistentProperty(association.foreignField());
                    // 添加外部关联
                    metadata.addCollection(targetEntity, property, accessor);
                } else {
                    // 都为空，从当前对象查找目标类型的reference，并自动添加关联
                    findForeignKey(targetEntity).ifPresent(property -> metadata.addCollection(targetEntity, property, accessor));
                }
            }
        }

        /**
         * 从当前实体查找目的实体的外键
         * 从当前实体找到关联目标实体的id
         *
         * @param entityClass 目的实体类型
         * @return 结果
         */
        private Optional<PersistentProperty<?>> findInnerKey(Class<?> entityClass) {
            return StreamSupport.stream(entity.getPersistentProperties(Reference.class).spliterator(), false)
                    .filter(property -> ClassUtils.isAssignable(entityClass, property.getRequiredAnnotation(Reference.class).value()))
                    .findFirst()
                    .map(p -> (PersistentProperty<?>) p);
        }

        /**
         * 查找目的实体的外键
         * 从目标实体中找到关联本实体的id
         *
         * @param targetEntity 目的实体信息
         * @return 结果
         */
        private Optional<PersistentProperty<?>> findForeignKey(PersistentEntity<?, ?> targetEntity) {
            Class<?> ownerEntityClass = entity.getType();
            return StreamSupport.stream(targetEntity.getPersistentProperties(Reference.class).spliterator(), false)
                    .filter(property -> ClassUtils.isAssignable(ownerEntityClass, property.getRequiredAnnotation(Reference.class).value()))
                    .findFirst()
                    .map(p -> (PersistentProperty<?>) p);
        }

    }
}
