package dev.flyfish.framework.dict.config;

import dev.flyfish.framework.annotations.DictValue;
import dev.flyfish.framework.dict.domain.Dictionary;
import dev.flyfish.framework.dict.domain.DictionaryValue;
import dev.flyfish.framework.dict.service.DictionaryService;
import dev.flyfish.framework.domain.base.Domain;
import dev.flyfish.framework.enums.BlankEnum;
import dev.flyfish.framework.enums.NamedEnum;
import dev.flyfish.framework.utils.Assert;
import dev.flyfish.framework.utils.ClassPathResourceScanner;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.stereotype.Component;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * 字典处理器
 *
 * @author wangyu
 * 用于反向注册默认值
 */
@RequiredArgsConstructor
@Component
@Slf4j
public class DictionaryProcessor implements InitializingBean {

    private final DictionaryService dictionaryService;

    @Override
    public void afterPropertiesSet() {
        String[] basePackages = DictionaryConfig.basePackages();
        Assert.notNull(basePackages, "未指定明确的字典扫描路径！");
        Stream<DictValue> annotations = ClassPathResourceScanner.forSuperType(Domain.class)
                .scan(basePackages)
                .flatMap(clazz -> Arrays.stream(FieldUtils.getFieldsWithAnnotation(clazz, DictValue.class)))
                .map(field -> field.getAnnotation(DictValue.class))
                .filter(annotation -> null != annotation && StringUtils.isNotBlank(annotation.value()) &&
                        BlankEnum.class != annotation.enumType());
        // 查找并更新
        Flux.fromStream(annotations)
                .distinct(DictValue::enumType)
                // 查找是否存在，不存在插入，存在无视
                .flatMap(annotation -> dictionaryService
                        .getByCode(annotation.value())
                        .map(dictionary -> applyDictProps(dictionary, annotation))
                        .switchIfEmpty(Mono.fromCallable(() -> applyDictProps(new Dictionary(), annotation)))
                )
                .collectList()
                .flatMapMany(dictionaryService::updateBatch)
                .subscribe();
    }

    /**
     * 给字典属性赋值
     *
     * @param dictionary 字典
     * @param annotation 注解
     * @return 结果
     */
    private Dictionary applyDictProps(Dictionary dictionary, DictValue annotation) {
        dictionary.setCode(annotation.value());
        if (CollectionUtils.isEmpty(dictionary.getValues())) {
            dictionary.setValues(mapValues(annotation.enumType()));
        }
        if (StringUtils.isBlank(dictionary.getDescription())) {
            dictionary.setDescription("系统自动创建的内置字典");
        }
        dictionary.setName(StringUtils.isNotBlank(annotation.name()) ? annotation.name() : annotation.value());
        return dictionary;
    }

    /**
     * 映射值，转换为字典表的值
     *
     * @return 结果
     */
    private List<DictionaryValue> mapValues(Class<? extends NamedEnum> enumClass) {
        return Arrays.stream(enumClass.getEnumConstants()).map(item -> {
            DictionaryValue value = new DictionaryValue();
            value.setEnable(true);
            value.setText(item.getName());
            value.setValue(((Enum<?>) item).name());
            return value;
        }).collect(Collectors.toList());
    }

}
